/*
 * This file performs all the Jewish to Julian and Gregorian date conversion.
 * Support for some observances is not included.  The algorithm for Rosh
 * Hashanah only goes back to about 300 on the Julian calendar.
 *
 * See "The comprehensive Hebrew calendar; its structure, history, and one
 * hundred years of corresponding dates: 5660-5760" by Arthur Spier.
 */

#include <stdio.h>
#include "structs.h"
#include "jmonths.h"
#include "defs.h"

#define DAYS_IN_400_JULIAN_YEARS	146100L
#define DAYS_IN_400_GREGORIAN_YEARS	146097L
#define JULIAN				(0)
#define GREGORIAN			(1)
#define DAYS_DIFFERENCE_IN_CALENDARS	(1373427L)
#define DAYS_TO_JULIAN_END		(639798L)	/* from (JAN, 1, 1) */

struct {
	struct date	date;
	long		days;
} Rosh_Hashanah, Next_Rosh_Hashanah;

long	days_to_Molad_Tishri(int year)
{
	int		postpone = 0;
	int		cycles;
	int		extra_years;
	int		months;
	long		days;
	int		parts;
	int		day;
	long		total_days;

	cycles = (year - 1) / YEARS_PER_CYCLE;
	extra_years = (year - 1) % YEARS_PER_CYCLE;
	months = cycles * MONTHS_PER_CYCLE + 12 * extra_years
	  + (extra_years + (extra_years > 6)) / 3;
	days = DAYS_PER_MONTH * months + TIME_0_DAY;
	parts = EXTRA_PARTS_PER_MONTH * months + TIME_0_PARTS;
	days += parts / PARTS_PER_DAY;
	parts %= PARTS_PER_DAY;
	total_days = days;
	day = days % DAYS_PER_WEEK + SUNDAY;

	if (parts > 18 * PARTS_PER_HOUR)
		postpone = 1;
	else if (!isjleap(year)) {
		if (day == TUESDAY && parts > 9 * PARTS_PER_HOUR + 204)
			postpone = 1;
		else if (isjleap(year - 1) && day == MONDAY
		  && parts > 15 * PARTS_PER_HOUR + 589)
			postpone = 1;
	}
	if (postpone) {
		day = (day + 1) % DAYS_PER_WEEK;
		total_days++;
	}
	if (day == SUNDAY || day == WEDNESDAY || day == FRIDAY) {
		day++;
		total_days++;
	}

	return total_days;
}

/* Julian day */
long		jday(struct date date, int calendar)
{
	int	century;
	int	extra_years;
	long	days_in_400_yrs;

	if (calendar == GREGORIAN)
		days_in_400_yrs = DAYS_IN_400_GREGORIAN_YEARS;
	else	/* JULIAN */
		days_in_400_yrs = DAYS_IN_400_JULIAN_YEARS;

	if (date.month > FEBRUARY)
		date.month -= 3;
	else {
		date.month += 9;
		date.year--;
	}

	century = date.year / 100;
	extra_years = date.year - 100 * century;

	return days_in_400_yrs * century / 4 + 1461 * extra_years / 4
	  + (153 * (date.month + 1 - JANUARY) + 2) / 5 + date.day;
}

/* Julian date */
struct date	jdate(long day, int calendar)
{
	struct date	date;
	long		days_in_400_yrs;

	if (calendar == GREGORIAN)
		days_in_400_yrs = DAYS_IN_400_GREGORIAN_YEARS;
	else	/* JULIAN */
		days_in_400_yrs = DAYS_IN_400_JULIAN_YEARS;

	date.year = (4 * day - 1) / days_in_400_yrs;
	day = 4 * day - 1 - days_in_400_yrs * date.year;
	date.day = day / 4;
	day = (4 * date.day + 3) / 1461;
	date.day = 4 * date.day + 3 - 1461 * day;
	date.day = (date.day + 4) / 4;
	date.month = (5 * date.day - 3) / 153;
	date.day = 5 * date.day - 3 - 153 * date.month;
	date.day = (date.day + 5) / 5;
	date.year = 100 * date.year + day;
	if (date.month < OCTOBER + (1 - JANUARY))
		date.month += 3;
	else {
		date.month -= 9;
		date.year++;
	}
	date.month += JANUARY - 1;

	return date;
}

struct date	gregorian_from_jewish(long daydiff)
{
	struct date	date1;
	struct date	date2;
	int		calendar;
	long		Julian_day;

	daydiff -= DAYS_DIFFERENCE_IN_CALENDARS;
	if (daydiff > DAYS_TO_JULIAN_END) {
		daydiff -= DAYS_TO_JULIAN_END;
		date1.day = FIRST_DAY_OF_GREGORIAN;
		date1.month = FIRST_MONTH_OF_GREGORIAN;
		date1.year = FIRST_YEAR_OF_GREGORIAN;
		calendar = GREGORIAN;
	} else {
		date1.day = date1.year = 1;
		date1.month = JANUARY;
		calendar = JULIAN;
	}

	Julian_day = jday(date1, calendar);
	date2 = jdate(Julian_day + daydiff, calendar);

	return date2;
}

/*
 * Determine when Rosh Hashanah occurred for the specified
 * Julian-Gregorian year and the next year.
 */
void	get_Roshes(int Gyear)
{
	int		Jyear;

	Jyear = Gyear + 3761;

	Rosh_Hashanah.days = days_to_Molad_Tishri(Jyear);
	Rosh_Hashanah.date = gregorian_from_jewish(Rosh_Hashanah.days);

	Next_Rosh_Hashanah.days = days_to_Molad_Tishri(Jyear + 1);
	Next_Rosh_Hashanah.date
	  = gregorian_from_jewish(Next_Rosh_Hashanah.days);
}

int	heshvan_is_long(int year_length)
{
	return year_length == 355 || year_length == 385;
}

int	kislev_is_short(int year_length)
{
	return year_length == 353 || year_length == 383;
}

void	change_odd_days(int *mon, int *day, int Jyear, int length)
{
	if (*mon == HESHVAN && *day == Jewish_month[HESHVAN].length + 1
	  && !heshvan_is_long(length)) {
		*mon = KISLEV;
		*day = 1;
	}
	if (*mon == KISLEV && *day == Jewish_month[KISLEV].length
	  && kislev_is_short(length)) {
		*mon = TEBETH;
		*day = 1;
	}
	if (*mon == ADAR && isjleap(Jyear))
		*mon = ADAR_II;
	else if (*mon == ADAR_II && !isjleap(Jyear))
		*mon = ADAR;
}

#define fj(m)			(29 * (m) + ((m) + 1) / 2)
#define jewish_to_ordinal(m)	((((m) + 6) % 12) + 1)

struct date	Getdate(int mon, int day, int year)
{
	static int	have_Roshes = 0;
	int		Jyear;
	int		Jyear_length;
	long		total_days;

	if (!have_Roshes || Rosh_Hashanah.date.year != year) {
		get_Roshes(year);
		have_Roshes = 1;
	}
	
	Jyear = year + 3761;

	Jyear_length = Next_Rosh_Hashanah.days - Rosh_Hashanah.days;
	change_odd_days(&mon, &day, Jyear, Jyear_length);

	total_days = Rosh_Hashanah.days;
	if (mon != TISHRI)
		total_days += fj(jewish_to_ordinal(mon) - 1);

	/* corrections for defective, excessive, and leap months */
	if (jewish_to_ordinal(mon) > jewish_to_ordinal(HESHVAN)
	  && heshvan_is_long(Jyear_length))
		total_days++;
	if (jewish_to_ordinal(mon) > jewish_to_ordinal(KISLEV)
	  && kislev_is_short(Jyear_length))
		total_days--;
	if (jewish_to_ordinal(mon) > jewish_to_ordinal(ADAR)
	  && isjleap(Jyear))
		total_days += 1 + Jewish_month[ADAR_II].length;

	total_days += day - 1;

	return gregorian_from_jewish(total_days);
}
